import numpy as np

class PostProcess:
  def __init__(self):
    pass

  def sortParam(self,var, mapMesh, cMeshDict, nFactor):

    # Post process delta star

    # Recompute deltaStar.
    var.blPar[9::var.nBlPar] = var.u[0::var.nVar] * var.u[1::var.nVar]

    # Map delta star to the coarse mesh and divide between the airfoil
    # and wake related points
    if mapMesh:

      deltaStar      = self.inverseMeshMapping(var.blPar[9::var.nBlPar], var.bNodes, cMeshDict['globCoord'], cMeshDict['bNodes'], nFactor)

      AFdeltaStarOut = deltaStar[:cMeshDict['bNodes'][2]]
      WKdeltaStarOut = deltaStar[cMeshDict['bNodes'][2]:]

    else:
      
      AFdeltaStarOut = var.blPar[9:var.bNodes[2]*var.nBlPar:var.nBlPar]
      WKdeltaStarOut = var.blPar[var.bNodes[2]*var.nBlPar + 9::var.nBlPar]

    # Compute blowing velocity on each cell.

    blwVelUp = np.zeros(var.bNodes[1] - 1)
    blwVelLw = np.zeros(var.bNodes[2] - var.bNodes[1])
    blwVelWk = np.zeros(var.nN - var.bNodes[2] - 1)

    # Upper
    i = 0
    for iPoint in range(1, var.bNodes[1]):

      iPrev = iPoint - 1

      blwVelUp[i] = (var.extPar[iPoint*var.nExtPar+1] * var.u[iPoint*var.nVar+3] * var.blPar[iPoint*var.nBlPar+9]
              - var.extPar[iPrev*var.nExtPar+1] * var.u[iPrev*var.nVar+3] * var.blPar[iPrev*var.nBlPar+9]) / (var.extPar[iPoint*var.nExtPar+1] * (var.xx[iPoint] - var.xx[iPrev]))

      i += 1

    # Lower
    i = 0
    for iPoint in range(var.bNodes[1], var.bNodes[2]):

      iPrev = 0 if iPoint == var.bNodes[1] else iPoint - 1

      blwVelLw[i] = (var.extPar[iPoint*var.nExtPar+1] * var.u[iPoint*var.nVar+3] * var.blPar[iPoint*var.nBlPar+9]
              - var.extPar[iPrev*var.nExtPar+1] * var.u[iPrev*var.nVar+3] * var.blPar[iPrev*var.nBlPar+9]) / (var.extPar[iPoint*var.nExtPar+1] * (var.xx[iPoint] - var.xx[iPrev]))

      i += 1

    # Wake
    i = 0
    for iPoint in range(var.bNodes[2] + 1, var.nN):

      iPrev = iPoint - 1

      blwVelWk[i] = (var.extPar[iPoint*var.nExtPar+1] * var.u[iPoint*var.nVar+3] * var.blPar[iPoint*var.nBlPar+9]
              - var.extPar[iPrev*var.nExtPar+1] * var.u[iPrev*var.nVar+3] * var.blPar[iPrev*var.nBlPar+9]) / (var.extPar[iPoint*var.nExtPar+1] * (var.xx[iPoint] - var.xx[iPrev]))

      i += 1

    if mapMesh:
      blwVelUp, blwVelLw, blwVelWk  = self.mapBlowingVelocities(blwVelUp, blwVelLw, blwVelWk, var.bNodes, cMeshDict['globCoord'], cMeshDict['bNodes'], var.xGlobal, nFactor)

    # Postprocess the boundary layer parameters.

    # Normalize friction and dissipation coefficients.
    frictionCoeff=var.u[3::var.nVar]**2 * var.blPar[2::var.nBlPar]
    dissipCoeff=var.u[3::var.nVar]**2 * var.blPar[1::var.nBlPar]

    # Compute total drag coefficient.
    CdTot=2*var.u[-5]*(var.u[-2]**((var.u[-4]+5)/2))

    # Compute friction and pressure drag coefficients. 
    Cdf_upper=np.trapz(frictionCoeff[:var.bNodes[1]],var.xGlobal[:var.bNodes[1]])
    # Insert stagnation point in the lower side. 
    Cdf_lower=np.trapz(np.insert(frictionCoeff[var.bNodes[1]:var.bNodes[2]],0,frictionCoeff[0]),
               np.insert(var.xGlobal[var.bNodes[1]:var.bNodes[2]],0,var.xGlobal[0]))
    Cdf=Cdf_upper+Cdf_lower # No friction in the wake
    Cdp=CdTot-Cdf

    # [CdTot, Cdf, Cdp]
    blScalars=[CdTot, Cdf, Cdp]

    # Store boundary layer final parameters in lists. 
    # [deltaStar, theta, H, Hk, Hstar, Hstar2, Cf, Cd, Ct, delta]
    blVectors_airfoil=[var.blPar[9:var.bNodes[2]*var.nBlPar:var.nBlPar],
              var.u[0:var.bNodes[2]*var.nVar:var.nVar],
              var.u[1:var.bNodes[2]*var.nVar:var.nVar],
              var.blPar[6:var.bNodes[2]*var.nBlPar:var.nBlPar],
              var.blPar[4:var.bNodes[2]*var.nBlPar:var.nBlPar],
              var.blPar[5:var.bNodes[2]*var.nBlPar:var.nBlPar],
              frictionCoeff[:var.bNodes[2]], dissipCoeff[:var.bNodes[2]],
              var.u[4:var.bNodes[2]*var.nVar:var.nVar],
              var.blPar[3:var.bNodes[2]*var.nBlPar:var.nBlPar]]

    blVectors_wake=[var.blPar[var.bNodes[2]*var.nBlPar+9::var.nBlPar],
            var.u[var.bNodes[2]*var.nVar+0::var.nVar],
            var.u[var.bNodes[2]*var.nVar+1::var.nVar],
            var.blPar[var.bNodes[2]*var.nBlPar+6::var.nBlPar],
            var.blPar[var.bNodes[2]*var.nBlPar+4::var.nBlPar],
            var.blPar[var.bNodes[2]*var.nBlPar+5::var.nBlPar],
            frictionCoeff[var.bNodes[2]:], dissipCoeff[var.bNodes[2]:],
            var.u[var.bNodes[2]*var.nVar+4::var.nVar],
            var.blPar[var.bNodes[2]*var.nBlPar+3::var.nBlPar]]

    return blScalars, blwVelUp, blwVelLw, blwVelWk, blVectors_airfoil, blVectors_wake, AFdeltaStarOut, WKdeltaStarOut

  def mapBlowingVelocities(self, blwVelUp, blwVelLw, blwVelWk, fbNodes, cMesh, cMeshbNodes, fMesh, nFactor):
    """Maps blowing velocties from cell to cell using weighted average interpolation.

        Args
        ----

        - blwVelUp (numpy.ndarray): Blowing velocities on the fine mesh of the upper side.

        - blwVelLw (numpy.ndarray): Blowing velocities on the fine mesh of the lower side.
    
        - blwVelWk (numpy.ndarray): Blowing velocities on the fine mesh of the wake.

        - fbNodes (numpy.ndarray): Boundary nodes of the fine mesh.

        - cMesh (numpy.ndarray): Coarse mesh coordinates in the global frame of reference.

        - cMeshbNodes (numpy.ndarray): Boundary nodes of the coarse mesh.

        - fMesh (numpy.ndarray): Fine mesh coordinates in the global frame of reference.

        - nFactor (int): Multiplicator underlying mesh adaptation.

    Example
    -------

    Fine mesh:   |-----------|-----------|-----------|----------|
               \	       /
           (1/2)	\	  	  / (1/2)
               \       /
                \     /
    Coarse mesh: |-----------------------|----------------------|
    """

    cblwVelUp = np.empty(len(cMesh[:cMeshbNodes[1]]) - 1)
    cblwVelLw = np.empty(len(cMesh[cMeshbNodes[1]:cMeshbNodes[2]]))
    cWKblwVel = np.empty(len(cMesh[cMeshbNodes[2]:]) - 1)

    for cCell in range(len(cblwVelUp)):

      cblwVelUp[cCell] = np.mean(blwVelUp[cCell*nFactor:(cCell + 1)*nFactor])
    
    for cCell in range(len(cblwVelLw)):

      cblwVelLw[cCell] = np.mean(blwVelLw[cCell*nFactor:(cCell + 1)*nFactor])

    for cCell in range(len(cWKblwVel)):

      cWKblwVel[cCell] = np.mean(blwVelWk[cCell*nFactor:(cCell + 1)*nFactor])

    return cblwVelUp, cblwVelLw, cWKblwVel


  def inverseMeshMapping(self, fVector, fbNodes, cMesh, cbNodes, nFactor):
    """Maps the variables required for blowing velocity computation from
    the fine mesh to the coarse mesh. Group separation is performed
    to avoid confusion between the points at the same place on the mesh
    but with different physical meaning (e.g First wake point).

    The mapping is based on a weigthed average with the nearest neighbors.

    Args
    ----

    - fVector (numpy.ndarray): Vector containing the function on the fine mesh.

    - fbNodes (numpy.ndarray): Id of the boundary nodes on the fine mesh.

    - cMesh (numpy.ndarray): Coarse mesh.

    - cbNodes (numpy.ndarray): Id of the boundary nodes on the coarse mesh.

    - nFactor (int): Multiplicator underlying mesh adaptation.

    Return
    ------

    - cVector (numpy.ndarray): Function contained in 'fVector' mapped to the coarse mesh.

    Example
    -------

    Fine mesh:   |-----------|-----------|-----------|----------|
                  \	         |          /
              (1/4)  \	     | (1/2)   / (1/4)
                  \        V        /
                    ----->   <-----
    Coarse mesh: |-----------------------|----------------------|
    """

    cVector = np.zeros(len(cMesh))

    for cPoint in range(len(cMesh)-1):
      cVector[cPoint] = fVector[cPoint * nFactor]
    cVector[-1] = fVector[-1]
    return cVector
