################################################################################
#                                                                              #
#                             Flow viscous module file                         #
#               Discretization: Computes solution gradient at one point        #
#                       of the computational domain. Different                 #
#                   finite difference spacial schemes are implemented.         #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
import numpy as np

class Discretization:
    """Discretize the spacial derivatives required to evaluate the boundary layer laws.
    
    Attributes
    ----------
    
    - xx: Cell coordinates in the (local) x-direction in the boundary layer.

    - xxExt: Cell coordinates in the (local) x-direction for the flow outside the boundary layer.

    - bNodes: Boundary nodes of the computational domain
    
    Returns
    -------
    
    Each method of the class returns the derivatives of the solution and of boundary layer parameters
    
    - du_dx: Derivative of the solution [dTheta/dx, dH/dx, dN/dx, due/dx, dCt/dx]

    - dHstar_dx: Derivative of Hstar

    - dueExt_dx, ddeltaStarExt_dx: External flow derivatives. Used in the interaction law.
    
    - c, cExt: Parameters dependent on the cell size 'dx', 'dxExt'. Used in the interation law.
    """
    def __init__(self, xx, xxExt, bNodes):
        dx=np.zeros(len(xx))
        for i in range(1,len(xx)):
            dx[i]=xx[i]-xx[0] if i==bNodes[1] else xx[i]-xx[i-1]
        self.dx=dx

        dxExt=np.zeros(len(xxExt))
        for i in range(1,len(xx)):
            dxExt[i]=xxExt[i]-xxExt[0] if i==bNodes[1] else xxExt[i]-xxExt[i-1]
        self.dxExt=dxExt

    def fou(self, var, x, idxPrev, idx):
        """First order backward difference.
        """
        du=x-var.u[var.nVar*idxPrev:var.nVar*(idxPrev+1)]

        dx = var.xx[idx] - var.xx[idxPrev]
        dxExt = var.extPar[idx*var.nExtPar + 4] - var.extPar[idxPrev*var.nExtPar + 4]
        
        du_dx=du/dx
        dHstar_dx=(var.blPar[idx*var.nBlPar+4]-var.blPar[idxPrev*var.nBlPar+4])/dx
        dueExt_dx=(var.extPar[idx*var.nExtPar+2]-var.extPar[idxPrev*var.nExtPar+2])/dxExt
        ddeltaStarExt_dx=(var.extPar[idx*var.nExtPar+3]-var.extPar[idxPrev*var.nExtPar+3])/dxExt
        c=4/(np.pi*dx)
        cExt=4/(np.pi*dxExt)
        return du_dx, dHstar_dx, dueExt_dx, ddeltaStarExt_dx, c, cExt
    
    def center(self, var, x, idxPrev2, idxPrev,idx, idxNext):
        """Second order central difference.
        """
        du= (var.u[idxNext*var.nVar:(idxNext+1)*var.nVar] - var.u[idxPrev*var.nVar:(idxPrev+1)*var.nVar])
        
        dx = var.xx[idx] - var.xx[idxPrev]
        dxExt = var.extPar[idx*var.nExtPar + 4] - var.extPar[idxPrev*var.nExtPar + 4]

        du_dx = du / (2*dx)
        dHstar_dx = (var.blPar[idxNext*var.nBlPar+4]-var.blPar[idxPrev*var.nBlPar+4]) / (2*dx)
        dueExt_dx = (var.extPar[idxNext*var.nExtPar+2] - var.extPar[idxPrev*var.nExtPar+2]) / (2*dxExt)
        ddeltaStarExt_dx=(var.extPar[idxNext*var.nExtPar+3]-var.extPar[idxPrev*var.nExtPar+3])/(2*dxExt)
        c=4/(np.pi*dx)
        cExt=4/(np.pi*dxExt)
        return du_dx, dHstar_dx, dueExt_dx, ddeltaStarExt_dx, c, cExt

    # Second order backward difference
    def sou(self, var, x,idxPrev2, idxPrev,idx, idxNext): # Attention if we want gradients at 2nd point after stagnation point on lower side. 
        """Second order upwing difference.
        """
        du=-3*x+4*var.u[+var.nVar*idxPrev:+var.nVar*(idxPrev+1)]-var.u[+var.nVar*idxPrev2:+var.nVar*(idxPrev2+1)]
        du_dx=du/(2*self.dx[idx])
        dHstar_dx=(3*var.blPar[idx*var.nBlPar+4]-4*var.blPar[idxPrev*var.nBlPar+4]+var.blPar[idxPrev2*var.nBlPar])/(2*self.dx[idx])
        dueExt_dx=(var.extPar[idxNext*var.nExtPar+2]-var.extPar[idxPrev*var.nExtPar+2])/(2*self.dx[idx])
        ddeltaStarExt_dx=(var.extPar[idxNext*var.nExtPar+3]-var.extPar[idxPrev*var.nExtPar+3])/(2*self.dx[idx])
        c=4/(np.pi*self.dx[idx])
        cExt=4/(np.pi*self.dxExt[idx])
        return du_dx, dHstar_dx, dueExt_dx, ddeltaStarExt_dx, c, cExt