#include "pSDSolver.h"
#include "pPanels.h"
#include <iostream>
#include <vector>
#include <cmath>

using namespace sdpm;

SDSolver::SDSolver(Panels *_pan, double _alpha, double _airspeed, unsigned long _maxIt)
{
    pan = _pan;
    fs_alpha = _alpha;
    fs_airspeed = _airspeed;
    maxIt = _maxIt;
}

void SDSolver::Run()
{   

    int R = 1; /* Wake type */

    Eigen::Vector3d fs_velocity;
    
    fs_velocity(0) = fs_airspeed * std::cos(fs_alpha);
    fs_velocity(1) = 0;
    fs_velocity(2) = fs_airspeed * std::sin(fs_alpha);

    double dt = pan->c0/pan->nPanChord/fs_velocity(0);

    pan->Xw.setZero(maxIt+1, pan->nPanSpan+1);
    pan->Yw.setZero(maxIt+1, pan->nPanSpan+1);
    pan->Zw.setZero(maxIt+1, pan->nPanSpan+1);

    Eigen::MatrixXd MAw = Eigen::MatrixXd::Zero(maxIt, pan->nPanSpan);
    pan->Xcw.setZero(maxIt, pan->nPanSpan);
    pan->Ycw.setZero(maxIt, pan->nPanSpan);
    pan->Zcw.setZero(maxIt, pan->nPanSpan);
    pan->nxw.setZero(maxIt, pan->nPanSpan);
    pan->nyw.setZero(maxIt, pan->nPanSpan);
    pan->nzw.setZero(maxIt, pan->nPanSpan);
    pan->tauxxw.setZero(maxIt, pan->nPanSpan);
    pan->tauxyw.setZero(maxIt, pan->nPanSpan);
    pan->tauxzw.setZero(maxIt, pan->nPanSpan);
    pan->tauyxw.setZero(maxIt, pan->nPanSpan);
    pan->tauyyw.setZero(maxIt, pan->nPanSpan);
    pan->tauyzw.setZero(maxIt, pan->nPanSpan);

    pan->ComputeInfluenceCoeffs_wing();

    pan->Xp0 = pan->Xp;
    pan->Yp0 = pan->Yp;
    pan->Zp0 = pan->Zp;

    /* std::cout << " xp yp zp " << std::endl;
    std::cout << pan->Xp << std::endl;
    std::cout << pan->Yp << std::endl;
    std::cout << pan->Zp << std::endl;

    std::cout << " xc yc zc " << std::endl;
    std::cout << pan->Xc << std::endl;
    std::cout << pan->Yc << std::endl;
    std::cout << pan->Zc << std::endl; */

    /* Calculate derivative of doublet strength in the chordwise direction */
    /* dX */
    Eigen::MatrixXd dXm1 = (pan->Xc.row(1)-pan->Xc.row(0));
    Eigen::MatrixXd dXm2 = (pan->Xc.block(2,0,2*pan->nPanChord,pan->Xc.cols()) - pan->Xc.block(0,0,2*pan->nPanChord-2,pan->Xc.cols()));
    Eigen::MatrixXd dXm3 = (pan->Xc.row(2*pan->nPanChord-1)-pan->Xc.row(2*pan->nPanChord-2));

    /* Vertical concatenation */
    Eigen::MatrixXd dXm(dXm1.rows()+dXm2.rows()+dXm3.rows(), dXm1.cols());
    dXm << dXm1,
            dXm2,
            dXm3;

    /* dY */
    Eigen::MatrixXd dYm1 = (pan->Yc.row(1)-pan->Yc.row(0));
    Eigen::MatrixXd dYm2 = (pan->Yc.block(2,0,2*pan->nPanChord,pan->Yc.cols()) - pan->Yc.block(0,0,2*pan->nPanChord-2,pan->Yc.cols()));
    Eigen::MatrixXd dYm3 = (pan->Yc.row(2*pan->nPanChord-1)-pan->Yc.row(2*pan->nPanChord-2));

    /* Vertical concatenation */
    Eigen::MatrixXd dYm(dYm1.rows()+dYm2.rows()+dYm3.rows(), dYm1.cols());
    dYm << dYm1,
            dYm2,
            dYm3;
    
    /* dZ */
    Eigen::MatrixXd dZm1 = (pan->Zc.row(1)-pan->Zc.row(0));
    Eigen::MatrixXd dZm2 = (pan->Zc.block(2,0,2*pan->nPanChord,pan->Zc.cols()) - pan->Zc.block(0,0,2*pan->nPanChord-2,pan->Zc.cols()));
    Eigen::MatrixXd dZm3 = (pan->Zc.row(2*pan->nPanChord-1)-pan->Zc.row(2*pan->nPanChord-2));

    /* Vertical concatenation */
    Eigen::MatrixXd dZm(dZm1.rows()+dZm2.rows()+dZm3.rows(), dZm1.cols());
    dZm << dZm1,
            dZm2,
            dZm3;
    
    Eigen::MatrixXd sm = (dXm.array()*dXm.array() + dYm.array()*dYm.array() + dZm.array()).sqrt();
    
    /* Chordwise tangent vector */
    Eigen::MatrixXd tmx = dXm.array()/sm.array();
    Eigen::MatrixXd tmy = dYm.array()/sm.array();
    Eigen::MatrixXd tmz = dZm.array()/sm.array();

    /* std::cout << " tm " << std::endl;
    std::cout << tmx << std::endl;
    std::cout << tmy << std::endl;
    std::cout << tmz << std::endl; */


    Eigen::MatrixXd dXn;
    Eigen::MatrixXd dYn;
    Eigen::MatrixXd dZn;

    if (pan->nPanSpan > 2)
    {
        /* Calculate derivative of doublet strength in the spanwise direction */
        /* dX */
        Eigen::MatrixXd dXn1 = (pan->Xc.col(1)-pan->Xc.col(0));
        Eigen::MatrixXd dXn2 = (pan->Xc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Xc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
        Eigen::MatrixXd dXn3 = (pan->Xc.col(pan->nPanSpan-1)-pan->Xc.col(pan->nPanSpan-2));

        /* Horizontal concatenation */
        dXn.resize(dXn1.rows(), dXn1.cols()+dXn2.cols()+dXn3.cols());
        dXn << dXn1, dXn2, dXn3;

        /* dY */
        Eigen::MatrixXd dYn1 = (pan->Yc.col(1)-pan->Yc.col(0));
        Eigen::MatrixXd dYn2 = (pan->Yc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Yc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
        Eigen::MatrixXd dYn3 = (pan->Yc.col(pan->nPanSpan-1)-pan->Yc.col(pan->nPanSpan-2));

        /* Horizontal concatenation */
        dYn.resize(dYn1.rows(), dYn1.cols()+dYn2.cols()+dYn3.cols());
        dYn << dYn1, dYn2, dYn3;

        /* dZ */
        Eigen::MatrixXd dZn1 = (pan->Zc.col(1)-pan->Zc.col(0));
        Eigen::MatrixXd dZn2 = (pan->Zc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Zc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
        Eigen::MatrixXd dZn3 = (pan->Zc.col(pan->nPanSpan-1)-pan->Zc.col(pan->nPanSpan-2));

        /* Horizontal concatenation */
        dZn.resize(dZn1.rows(), dZn1.cols()+dZn2.cols()+dZn3.cols());
        dZn << dZn1, dZn2, dZn3;

    }
    else if (pan->nPanSpan == 2){

        /* Calculate derivative of doublet strength in the spanwise direction */
        /* dX */
        Eigen::MatrixXd dXn1 = (pan->Xc.col(1)-pan->Xc.col(0));

        /* Horizontal concatenation */
        dXn.resize(dXn1.rows(), dXn1.cols()+dXn1.cols());
        dXn << dXn1, dXn1;

        /* dY */
        Eigen::MatrixXd dYn1 = (pan->Yc.col(1)-pan->Yc.col(0));

        /* Horizontal concatenation */
        dYn.resize(dYn1.rows(), dYn1.cols()+dYn1.cols());
        dYn << dYn1, dYn1;

        /* dZ */
        Eigen::MatrixXd dZn1 = (pan->Zc.col(1)-pan->Zc.col(0));

        /* Horizontal concatenation */
        dZn.resize(dZn1.rows(), dZn1.cols()+dZn1.cols());
        dZn << dZn1, dZn1;

    }
    else{
        throw;
    }

    Eigen::MatrixXd sn = (dXn.array()*dXn.array() + dYn.array()*dYn.array() + dZn.array()).sqrt();

    /* std::cout << " dxm dym dzm " << std::endl;
    std::cout << dXm << std::endl;
    std::cout << dYm << std::endl;
    std::cout << dZm << std::endl;

    std::cout << " dxn dyn dzn " << std::endl;
    std::cout << dXn << std::endl;
    std::cout << dYn << std::endl;
    std::cout << dZn << std::endl; */


    /* Calculate spanwise tangent vector */

    Eigen::MatrixXd tnx = dXn.array()/sn.array();
    Eigen::MatrixXd tny = dYn.array()/sn.array();
    Eigen::MatrixXd tnz = dZn.array()/sn.array();

    /* std::cout << " tnx tny tnz " << std::endl;
    std::cout << tnx << std::endl;
    std::cout << tny << std::endl;
    std::cout << tnz << std::endl; */

    /* Calculate source strengths */

    Eigen::MatrixXd SA = -(fs_velocity[0]*pan->nx + fs_velocity[1]*pan->ny + fs_velocity[2]*pan->nz);
    Eigen::MatrixXd dum = SA.transpose(); 
    Eigen::MatrixXd sigma = dum.reshaped(2*pan->nPanSpan*pan->nPanChord,1);

    /* std::cout << " SA sigma " << std::endl;
    std::cout << SA << std::endl;
    std::cout << sigma << std::endl; */

    /* Solution at zero time */

    Eigen::MatrixXd LHS = pan->BN;
    Eigen::MatrixXd RHS = -pan->AN * sigma;
    Eigen::VectorXd Sol = LHS.partialPivLu().solve(RHS);

    Eigen::MatrixXd Phi0 = Sol.reshaped(pan->nPanSpan, 2*pan->nPanChord).transpose();

    /* std::cout << "Phi0 " << Phi0 << std::endl; */

    /* First wake row */
    pan->Xw.row(0) = pan->Xp.row(2*pan->nPanChord);
    pan->Yw.row(0) = pan->Yp.row(2*pan->nPanChord);
    pan->Zw.row(0) = pan->Zp.row(2*pan->nPanChord);

    CL.setZero(maxIt);
    CD.setZero(maxIt);
    CY.setZero(maxIt);

    std::vector<Eigen::MatrixXd> Phi(maxIt);

    unsigned long it = 1;

    while(it < maxIt){

        std::cout << it << std::endl;

        /* Propagate wake */
        Eigen::MatrixXd uwp = Eigen::MatrixXd::Zero(it-1, pan->nPanSpan + 1);
        Eigen::MatrixXd vwp = Eigen::MatrixXd::Zero(it-1, pan->nPanSpan + 1);
        Eigen::MatrixXd wwp = Eigen::MatrixXd::Zero(it-1, pan->nPanSpan + 1);

        Eigen::MatrixXd dum0(1+uwp.rows(), uwp.cols());
        dum0 << Eigen::MatrixXd::Zero(1,pan->nPanSpan+1),
                uwp;
        
        pan->Xw.block(1,0,it,pan->Xw.cols()) = pan->Xw.block(0,0,it-1,pan->Xw.cols()) + (Eigen::MatrixXd::Constant(dum0.rows(), dum0.cols(), fs_velocity(0)) + dum0)*dt;

        Eigen::MatrixXd dum1(1+vwp.rows(), vwp.cols());
        dum1 << Eigen::MatrixXd::Zero(1,pan->nPanSpan+1),
                vwp;
        pan->Yw.block(1,0,it,pan->Yw.cols()) = pan->Yw.block(0,0,it-1,pan->Yw.cols()) + (Eigen::MatrixXd::Constant(dum1.rows(), dum1.cols(), fs_velocity(1)) + dum1)*dt;

        Eigen::MatrixXd dum2(1+uwp.rows(), uwp.cols());
        dum2 << Eigen::MatrixXd::Zero(1,pan->nPanSpan+1),
                wwp;
        pan->Zw.block(1,0,it,pan->Zw.cols()) = pan->Zw.block(0,0,it-1,pan->Zw.cols()) + (Eigen::MatrixXd::Constant(dum2.rows(), dum2.cols(), fs_velocity(2)) + dum2)*dt;

        if (it > 1){
            MAw.block(1,0,it ,MAw.cols()) = MAw.block(0,0,it-1,MAw.cols());
        }

        pan->Xw.row(0) = pan->Xp.row(2*pan->nPanChord);
        pan->Yw.row(0) = pan->Yp.row(2*pan->nPanChord);
        pan->Zw.row(0) = pan->Zp.row(2*pan->nPanChord);

        /* Calculate wake control points */

        pan->WakeControlPoints(it);

        /* Calculate wake influence coefficients */

        pan->ComputeInfluenceCoeffs_wake(it);

        /* Set up Kutta condition */

        pan->SetKutta();

        /* Equations */

        Eigen::MatrixXd LHS;
        
        /* std::cout << " Here " << std::endl;
        std::cout << pan->CN.block(0,0,pan->CN.rows(), pan->nPanSpan) << std::endl; */

        /* Horizontal concatenation */
        Eigen::MatrixXd LHS_dum1(pan->BN.rows(), pan->BN.cols()+pan->CN.block(0,0,pan->CN.rows(), pan->nPanSpan).cols());
        LHS_dum1 << pan->BN, pan->CN.block(0,0,pan->CN.rows(), pan->nPanSpan);

        /* Vertical concatenation */
        LHS.setZero(LHS_dum1.rows()+pan->Kutta.rows(), LHS_dum1.cols());
        LHS << LHS_dum1,
            pan->Kutta;

        Eigen::MatrixXd RHS;

        if (it==1){
            Eigen::MatrixXd RHS_dum1;
            RHS_dum1 = -pan->AN*sigma;

            RHS.setZero(RHS_dum1.rows()+Eigen::MatrixXd::Zero(pan->nPanSpan, 1).rows(), RHS_dum1.cols());
            RHS << RHS_dum1,
                Eigen::MatrixXd::Zero(pan->nPanSpan, 1);
        }
        else{
            
            Eigen::MatrixXd RHS_dum1;
            RHS_dum1 = -pan->AN*sigma - pan->CN.block(0,pan->nPanSpan, pan->CN.rows(), (it-1)*pan->nPanSpan) * MAw.block(1,0,it-1,MAw.cols()).transpose().reshaped((it-1)*pan->nPanSpan,1);

            RHS.setZero(RHS_dum1.rows()+Eigen::MatrixXd::Zero(pan->nPanSpan, 1).rows(), RHS_dum1.cols());
            RHS << RHS_dum1,
                Eigen::MatrixXd::Zero(pan->nPanSpan, 1);

        }

        /* Solve for strength */

        /* std::cout << " LHS " << std::endl;
        std::cout << LHS << std::endl;

        std::cout << " RHS " << std::endl;
        std::cout << RHS << std::endl; */

        Eigen::VectorXd Mu2;
        Mu2.setZero(RHS.size());
        Mu2 = LHS.partialPivLu().solve(RHS);

        /* std::cout << " Mu2 " << std::endl;
        std::cout << Mu2 << std::endl; */

        /*std::cout << "mu2" << std::endl;
        std::cout << Mu2 << std::endl; */

        Eigen::MatrixXd MA_dum1 = Mu2.reshaped(pan->nPanSpan, 2*pan->nPanChord+1).transpose();

        /* std::cout << "madum" << std::endl;
        std::cout << MA_dum1 << std::endl; */

        MAw.row(0) = MA_dum1.row(2*pan->nPanChord);

        /* std::cout << "maw" << std::endl;
        std::cout << MAw << std::endl; */

        Eigen::VectorXd Mu(2*pan->nPanChord*pan->nPanSpan);
        

        for (size_t i=0; i<2*pan->nPanChord*pan->nPanSpan; ++i)
        {
            Mu(i) = Mu2(i);
        }
        /* std::cout << "mu" << std::endl;
        std::cout << Mu << std::endl; */

        Eigen::MatrixXd MA(2*pan->nPanChord, MA_dum1.cols());

        for (size_t i=0; i<2*pan->nPanChord; ++i)
        {
            MA.row(i) = MA_dum1.row(i);
        }
        
        /* std::cout << "ma" << std::endl;
        std::cout << MA << std::endl; */

        Phi[it] = MA;

        /* std::cout << " Phi[it] " << Phi[it] << std::endl; */

        /* std::cout << " phi it " << std::endl;
        std::cout << Phi[it] << std::endl; */

        /* dMA_chord */
        Eigen::MatrixXd dMA1_chord = (MA.row(1)-MA.row(0));
        Eigen::MatrixXd dMA2_chord = (MA.block(2,0,2*pan->nPanChord-1,MA.cols()) - MA.block(0,0,2*pan->nPanChord-2,MA.cols()));
        Eigen::MatrixXd dMA3_chord = (MA.row(2*pan->nPanChord-1)-MA.row(2*pan->nPanChord-2));

        /* Vertical concatenation */
        Eigen::MatrixXd dMA_chord(dMA1_chord.rows()+dMA2_chord.rows()+dMA3_chord.rows(), dMA1_chord.cols());
        dMA_chord << dMA1_chord,
                     dMA2_chord,
                     dMA3_chord;
        
        Eigen::MatrixXd um = dMA_chord.array()/sm.array();

        /* dMA_span */

        Eigen::MatrixXd dMA_span;

        if (pan->nPanSpan > 2){
            Eigen::MatrixXd dMA1_span = (MA.col(1)-MA.col(0));
            Eigen::MatrixXd dMA2_span = (MA.block(0,2,2*pan->nPanChord,pan->nPanSpan) - MA.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
            Eigen::MatrixXd dMA3_span = (MA.col(pan->nPanSpan-1)-MA.col(pan->nPanSpan-2));

            /* Horizontal concatenation */
            dMA_span.setZero(dMA1_span.rows(), dMA1_span.cols()+dMA2_span.cols()+dMA3_span.cols());
            dMA_span << dMA1_span, dMA2_span, dMA3_span;
        }
        else if (pan->nPanSpan == 2){
            Eigen::MatrixXd dMA1_span = (MA.col(1)-MA.col(0));

            /* Horizontal concatenation */
            dMA_span.setZero(dMA1_span.rows(), dMA1_span.cols()+dMA1_span.cols());
            dMA_span << dMA1_span, dMA1_span;
        }
        else{
            throw;
        }

        Eigen::MatrixXd un = dMA_span.array()/sn.array();

        /* Calculate velocities on control points in the x,y,z directions
           On each control point we project the velocities in the chordwise,
           spanwise and normal directions to the x,y,z axes */
        Eigen::MatrixXd uc(2*pan->nPanChord, pan->nPanSpan);
        Eigen::MatrixXd vc(2*pan->nPanChord, pan->nPanSpan);
        Eigen::MatrixXd wc(2*pan->nPanChord, pan->nPanSpan);

        Eigen::Matrix3d Rmat_sys;
        Eigen::Vector3d RHS_sys;
        Eigen::VectorXd dummy_sys;

        for (size_t i=0;i<2*pan->nPanChord; ++i)
        {
            for (size_t j=0; j<pan->nPanSpan; ++j)
            {
                Rmat_sys(0,0) = tmx(i,j);
                Rmat_sys(0,1) = tmy(i,j);
                Rmat_sys(0,2) = tmz(i,j);

                Rmat_sys(1,0) = tnx(i,j);
                Rmat_sys(1,1) = tny(i,j);
                Rmat_sys(1,2) = tnz(i,j);

                Rmat_sys(2,0) = pan->nx(i,j);
                Rmat_sys(2,1) = pan->ny(i,j);
                Rmat_sys(2,2) = pan->nz(i,j);

                RHS_sys(0) = um(i,j);
                RHS_sys(1) = un(i,j);
                RHS_sys(2) = SA(i,j);

                dummy_sys = Rmat_sys.partialPivLu().solve(RHS_sys);
                uc(i,j)=dummy_sys(0);
                vc(i,j)=dummy_sys(1);
                wc(i,j)=dummy_sys(2);
            }
        }

        uc = Eigen::MatrixXd::Constant(uc.rows(), uc.cols(), fs_velocity(0)) + uc;
        vc = Eigen::MatrixXd::Constant(vc.rows(), vc.cols(), fs_velocity(1)) + vc;
        wc = Eigen::MatrixXd::Constant(wc.rows(), wc.cols(), fs_velocity(2)) + wc;

        /* Calculate presssure coefficient */
        Eigen::MatrixXd dPhidt;
        if (it==1){
            Eigen::Map<Eigen::VectorXd> Phi1Flatten(Phi[1].data(),Phi[1].size());
            Eigen::Map<Eigen::VectorXd> Phi0Flatten(Phi0.data(),Phi0.size());
            dPhidt = (Phi1Flatten-Phi0Flatten)/dt;
        }
        else{
            Eigen::Map<Eigen::VectorXd> Phi1Flatten(Phi[it].data(),Phi[it].size());
            Eigen::Map<Eigen::VectorXd> Phi0Flatten(Phi[it-1].data(),Phi[it-1].size());
            dPhidt = (Phi1Flatten - Phi0Flatten)/dt;
        }

        Eigen::MatrixXd cp_dum = (uc.array()*uc.array() + vc.array()*vc.array() + wc.array()*wc.array())*1/fs_airspeed/fs_airspeed - 2*dPhidt.array()*1/fs_airspeed/fs_airspeed;
        cp = Eigen::MatrixXd::Constant(uc.rows(), uc.cols(), 1) - cp_dum;


        /* Forces on panels */

        std::cout << "cp" << std::endl;
        std::cout << cp << std::endl;
        std::cout << "pan->s" << std::endl;
        std::cout << pan->s << std::endl;
        std::cout << "pan->nx" << std::endl;
        std::cout << pan->nx << std::endl;
        std::cout << "pan->Surface" << std::endl;
        std::cout << pan->Surface << std::endl;

        Eigen::MatrixXd fx = -cp.array()*pan->s.array()*pan->nx.array()/pan->Surface;
        Eigen::MatrixXd fy = -cp.array()*pan->s.array()*pan->nx.array()/pan->Surface;
        Eigen::MatrixXd fz = -cp.array()*pan->s.array()*pan->nx.array()/pan->Surface;

        std::cout << fx << std::endl;
        std::cout << fy << std::endl;
        std::cout << fz << std::endl;

        double Fx = fx.sum();
        //double Fy = fy.sum();
        double Fz = fz.sum();

        if (R>0){
            CL(it) = Fz;
            CD(it) = Fx;
        }
        else{
            CL(it) = Fz-Fx*fs_alpha;
            CD(it) = Fz*fs_alpha+Fx;
        }
        std::cout << " CL = " << CL(it) << std::endl;
        it++;

    }
}


    // int nChords = 100;
    // double dt = nChords * pan->Getc0() / fs_velocity[0];

    // /* Source strength */

    // Eigen::MatrixXd SA = -(fs_velocity[0]*pan->nx/fs_beta + fs_velocity[1]*pan->ny + fs_velocity[2]*pan->nz);
    // Eigen::MatrixXd dum = SA.transpose(); 
    // Eigen::MatrixXd sigma = dum.reshaped(2*pan->nPanSpan*pan->nPanChord,1);

    // pan->CreateWake(fs_velocity[0], fs_velocity[1], fs_velocity[2], fs_beta, dt);
    // std::cout << "Wake created." << std::endl;

    // pan->ComputeInfluenceCoeffs_wing();
    // pan->ComputeInfluenceCoeffs_wake();
    // std::cout << "Computed influence coefficients." << std::endl;

    // pan->SetKutta();
    // std::cout << "Imposed Kutta condition." << std::endl;

    // /* Set up Dirichlet boundary condition */

    // /* Horizontal concatenation */
    // Eigen::MatrixXd LHS_dum1(pan->BN.rows(), pan->BN.cols()+pan->CN.cols());
    // LHS_dum1 << pan->BN, pan->CN;

    // /* Vertical concatenation */
    // Eigen::MatrixXd LHS(LHS_dum1.rows()+pan->Kutta.rows(), LHS_dum1.cols());
    // LHS << LHS_dum1,
    //         pan->Kutta;

    // Eigen::MatrixXd RHS_dum1;
    // RHS_dum1 = -pan->AN*sigma;

    // Eigen::MatrixXd RHS(RHS_dum1.rows()+Eigen::MatrixXd::Zero(pan->nPanSpan, 1).rows(), RHS_dum1.cols());
    // RHS << RHS_dum1,
    //     Eigen::MatrixXd::Zero(pan->nPanSpan, 1);

    // /* Solve for strength */
    // Eigen::VectorXd Mu = LHS.partialPivLu().solve(RHS);

    // Eigen::MatrixXd MA_dum1;
    // MA_dum1 = Mu.reshaped(pan->nPanSpan, 2*pan->nPanChord+1).transpose();

    // /* Wake doublet strengths */

    // Eigen::RowVectorXd Mw = MA_dum1.row(MA_dum1.rows()-1); // Last row of MA_dum1

    // Eigen::VectorXd Mu2(2*pan->nPanChord*pan->nPanSpan);

    // for (size_t i=0; i<2*pan->nPanChord*pan->nPanSpan; ++i)
    // {
    //     Mu2.row(i) = Mu.row(i);
    // }

    // Eigen::MatrixXd MA(2*pan->nPanChord, MA_dum1.cols());

    // for (size_t i=0; i<2*pan->nPanChord; ++i)
    // {
    //     MA.row(i) = MA_dum1.row(i);
    // }

    // /* Calculate derivative of doublet strength in the chordwise direction */
    // /* dX */
    // Eigen::MatrixXd dXc1_chord = (pan->Xc.row(1)-pan->Xc.row(0));
    // Eigen::MatrixXd dXc2_chord = (pan->Xc.block(2,0,2*pan->nPanChord,pan->Xc.cols()) - pan->Xc.block(0,0,2*pan->nPanChord-2,pan->Xc.cols()));
    // Eigen::MatrixXd dXc3_chord = (pan->Xc.row(2*pan->nPanChord-1)-pan->Xc.row(2*pan->nPanChord-2));

    // /* Vertical concatenation */
    // Eigen::MatrixXd dXc_chord(dXc1_chord.rows()+dXc2_chord.rows()+dXc3_chord.rows(), dXc1_chord.cols());
    // dXc_chord << dXc1_chord,
    //         dXc2_chord,
    //         dXc3_chord;

    // /* dY */
    // Eigen::MatrixXd dYc1_chord= (pan->Yc.row(1)-pan->Yc.row(0));
    // Eigen::MatrixXd dYc2_chord= (pan->Yc.block(2,0,2*pan->nPanChord,pan->Yc.cols()) - pan->Yc.block(0,0,2*pan->nPanChord-2,pan->Yc.cols()));
    // Eigen::MatrixXd dYc3_chord = (pan->Yc.row(2*pan->nPanChord-1)-pan->Yc.row(2*pan->nPanChord-2));

    // /* Vertical concatenation */
    // Eigen::MatrixXd dYc_chord(dYc1_chord.rows()+dYc2_chord.rows()+dYc3_chord.rows(), dYc1_chord.cols());
    // dYc_chord << dYc1_chord,
    //         dYc2_chord,
    //         dYc3_chord;

    // /* dZ */
    // Eigen::MatrixXd dZc1_chord = (pan->Zc.row(1)-pan->Zc.row(0));
    // Eigen::MatrixXd dZc2_chord = (pan->Zc.block(2,0,2*pan->nPanChord,pan->Zc.cols()) - pan->Zc.block(0,0,2*pan->nPanChord-2,pan->Zc.cols()));
    // Eigen::MatrixXd dZc3_chord = (pan->Zc.row(2*pan->nPanChord-1)-pan->Zc.row(2*pan->nPanChord-2));

    // /* Vertical concatenation */
    // Eigen::MatrixXd dZc_chord(dZc1_chord.rows()+dZc2_chord.rows()+dZc3_chord.rows(), dZc1_chord.cols());
    // dZc_chord << dZc1_chord,
    //             dZc2_chord,
    //             dZc3_chord;
    
    // /* dMA_chord */
    // Eigen::MatrixXd dMA1_chord = (MA.row(1)-MA.row(0));
    // Eigen::MatrixXd dMA2_chord = (MA.block(2,0,2*pan->nPanChord,MA.cols()) - MA.block(0,0,2*pan->nPanChord-2,MA.cols()));
    // Eigen::MatrixXd dMA3_chord = (MA.row(2*pan->nPanChord-1)-MA.row(2*pan->nPanChord-2));

    // /* Vertical concatenation */
    // Eigen::MatrixXd dMA_chord(dMA1_chord.rows()+dMA2_chord.rows()+dMA3_chord.rows(), dMA1_chord.cols());
    // dMA_chord << dMA1_chord,
    //             dMA2_chord,
    //             dMA3_chord;

    // /* Calculate chordwise tangent vectors */
    // /* .array() member of MatrixXd allows the structure to be treated as ArrayXd for the operation.
    // This allows product element by element instead of the classic matrix product of MatrixXd (aliasing is present for ArrayXd*ArrayXd)*/
    // Eigen::MatrixXd sc_chord = (dXc_chord.array()*dXc_chord.array()+dYc_chord.array()*dYc_chord.array()+dZc_chord.array()*dZc_chord.array()).sqrt();

    // Eigen::MatrixXd tmx_chord = dXc_chord.array()/sc_chord.array();
    // Eigen::MatrixXd tmy_chord = dYc_chord.array()/sc_chord.array();
    // Eigen::MatrixXd tmz_chord = dZc_chord.array()/sc_chord.array();
    // /* Calculate the velocity in the chordwise direction */
    // Eigen::MatrixXd um = dMA_chord.array()/sc_chord.array();
    
    
    // /* Calculate derivative of doublet strength in the spanwise direction */
    // /* dX */
    // Eigen::MatrixXd dXc1_span = (pan->Xc.col(1)-pan->Xc.col(0));
    // Eigen::MatrixXd dXc2_span = (pan->Xc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Xc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
    // Eigen::MatrixXd dXc3_span = (pan->Xc.col(pan->nPanSpan-1)-pan->Xc.col(pan->nPanSpan-2));

    // /* Horizontal concatenation */
    // Eigen::MatrixXd dXc_span(dXc1_span.rows(), dXc1_span.cols()+dXc2_span.cols()+dXc3_span.cols());
    // dXc_span << dXc1_span, dXc2_span, dXc3_span;

    // /* dY */
    // Eigen::MatrixXd dYc1_span = (pan->Yc.col(1)-pan->Yc.col(0));
    // Eigen::MatrixXd dYc2_span = (pan->Yc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Yc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
    // Eigen::MatrixXd dYc3_span = (pan->Yc.col(pan->nPanSpan-1)-pan->Yc.col(pan->nPanSpan-2));

    // /* Horizontal concatenation */
    // Eigen::MatrixXd dYc_span(dYc1_span.rows(), dYc1_span.cols()+dYc2_span.cols()+dYc3_span.cols());
    // dYc_span << dYc1_span, dYc2_span, dYc3_span;

    // /* dZ */
    // Eigen::MatrixXd dZc1_span = (pan->Zc.col(1)-pan->Zc.col(0));
    // Eigen::MatrixXd dZc2_span = (pan->Zc.block(0,2,2*pan->nPanChord,pan->nPanSpan) - pan->Zc.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
    // Eigen::MatrixXd dZc3_span = (pan->Zc.col(pan->nPanSpan-1)-pan->Zc.col(pan->nPanSpan-2));

    // /* Horizontal concatenation */
    // Eigen::MatrixXd dZc_span(dZc1_span.rows(), dZc1_span.cols()+dZc2_span.cols()+dZc3_span.cols());
    // dZc_span << dZc1_span, dZc2_span, dZc3_span;
    
    // /* dMA_span */
    // Eigen::MatrixXd dMA1_span = (MA.col(1)-MA.col(0));
    // Eigen::MatrixXd dMA2_span = (MA.block(0,2,2*pan->nPanChord,pan->nPanSpan) - MA.block(0,0,2*pan->nPanChord,pan->nPanSpan-2));
    // Eigen::MatrixXd dMA3_span = (MA.col(pan->nPanSpan-1)-MA.col(pan->nPanSpan-2));

    // /* Horizontal concatenation */
    // Eigen::MatrixXd dMA_span(dMA1_span.rows(), dMA1_span.cols()+dMA2_span.cols()+dMA3_span.cols());
    // dMA_span << dMA1_span, dMA2_span, dMA3_span;

    // /* Calculate spanwise tangent vectors */
    // /* .array() member of MatrixXd allows the structure to be treated as ArrayXd for the operation.
    // This allows product element by element instead of the classic matrix product of MatrixXd (aliasing is present for ArrayXd*ArrayXd)*/
    // Eigen::MatrixXd sc_span = (dXc_span.array()*dXc_span.array()+dYc_span.array()*dYc_span.array()+dZc_span.array()*dZc_span.array()).sqrt();

    // Eigen::MatrixXd tnx_span = dXc_span.array()/sc_span.array();
    // Eigen::MatrixXd tny_span = dYc_span.array()/sc_span.array();
    // Eigen::MatrixXd tnz_span = dZc_span.array()/sc_span.array();
    // /* Calculate the velocity in the chordwise direction */
    // Eigen::MatrixXd un = dMA_span.array()/sc_span.array();

    // /* Calculate velocities on control points in the x,y,z directions
    //    On each control point we project the velocities in the chordwise,
    //    spanwise and normal directions to the x,y,z axes */
    // Eigen::MatrixXd uc(2*pan->nPanChord, pan->nPanSpan);
    // Eigen::MatrixXd vc(2*pan->nPanChord, pan->nPanSpan);
    // Eigen::MatrixXd wc(2*pan->nPanChord, pan->nPanSpan);

    // Eigen::Matrix3d Rmat_sys;
    // Eigen::Vector3d RHS_sys;
    // Eigen::VectorXd dummy_sys;

    // for (size_t i=0;i<2*pan->nPanChord; ++i)
    // {
    //     for (size_t j=0; j<pan->nPanSpan; ++j)
    //     {
    //         Rmat_sys(0,0) = tmx_chord(i,j);
    //         Rmat_sys(0,1) = tmy_chord(i,j);
    //         Rmat_sys(0,2) = tmz_chord(i,j);

    //         Rmat_sys(1,0) = tnx_span(i,j);
    //         Rmat_sys(1,1) = tny_span(i,j);
    //         Rmat_sys(1,2) = tnz_span(i,j);

    //         Rmat_sys(2,0) = pan->nx(i,j);
    //         Rmat_sys(2,1) = pan->ny(i,j);
    //         Rmat_sys(2,2) = pan->nz(i,j);

    //         RHS_sys(0) = um(i,j);
    //         RHS_sys(1) = un(i,j);
    //         RHS_sys(2) = SA(i,j); // or SA0?

    //         dummy_sys = Rmat_sys.partialPivLu().solve(RHS_sys);
    //         uc(i,j)=dummy_sys(0);
    //         vc(i,j)=dummy_sys(1);
    //         wc(i,j)=dummy_sys(2);
    //     }
    // }

    // /* Scale velocity to compressible space */
    // uc/=fs_beta;
    // uc+=Eigen::MatrixXd::Constant(uc.rows(),uc.cols(),fs_velocity(0));
    // vc+=Eigen::MatrixXd::Constant(vc.rows(),vc.cols(),fs_velocity(1));
    // wc+=Eigen::MatrixXd::Constant(wc.rows(),wc.cols(),fs_velocity(2));

    // std::cout << " uc " << uc << std::endl;
    // std::cout << " vc " << vc << std::endl;
    // std::cout << " wc " << wc << std::endl;


    // /* Calculate pressure coefficient using second order relationship */
    // Eigen::MatrixXd bracket = Eigen::MatrixXd::Constant(uc.rows(),uc.cols(),1/2).array()*(uc.array()*uc.array()+vc.array()*vc.array()+wc.array()*wc.array()) - Eigen::MatrixXd::Constant(uc.rows(),uc.cols(),1/2*fs_airspeed*fs_airspeed).array();

    // std::cout << "brack " << bracket << std::endl;

    // Eigen::MatrixXd cp=(-bracket.array() + Eigen::MatrixXd::Constant(bracket.rows(), bracket.cols(),1/2/(fs_asound*fs_asound)).array() * bracket.array() * bracket.array()) * Eigen::MatrixXd::Constant(bracket.rows(), bracket.cols(),2/(fs_airspeed*fs_airspeed)).array();
    // std::cout << " cp " << cp << std::endl;
    
    // % Calculate total velocity
    // Vtot2=uc.^2+vc.^2+wc.^2;
    
    // % Calculate maximum velocity (giving Mach=1);
    // Vmax=Q*sqrt(1+2/((gama-1)*Mach^2));
    // % Calculate local speed of sound
    // asound_local=asound*sqrt(1-(gama-1)/2*Mach^2*(Vtot2/Q^2-1));
    // % Calculate local Mach number
    // Mach_local=sqrt(Vtot2)./asound_local;
    // % Calculate critical velocity (where local Mach number = 1)
    // Vcrit=sqrt((gama-1)/(gama+1))*Vmax;
    // iko=find(sqrt(Vtot2) >= Vcrit);
    // if ~isempty(iko)
    //     disp('Warning! Sonic Mach number exceeded')
    //     super=1;
    // end
    // supervec(ialpha)=super;
    
    // % Calculate forces on panels
    // fx=-cp.*s0.*nx0/S;
    // fy=-cp.*s0.*ny0/S;
    // fz=-cp.*s0.*nz0/S;
    
    // % Calculate total forces on wing in the x,y,z directions
    // Fx=sum(sum(fx));
    // Fy=sum(sum(fy));
    // Fy_half=sum(sum(fy(:,n/2+1:n)));
    // Fz=sum(sum(fz));
    
    // % Calculate lift and drag coefficients
    // CL(ialpha)=Fz*cos(alpha)-Fx*sin(alpha);
    // CD(ialpha)=Fz*sin(alpha)+Fx*cos(alpha);
    // CY(ialpha)=Fy;
    // CY_half(ialpha)=Fy_half;
    
    // % Calculate lift and drag distributions
    // dy=Yp(1,2:n+1)-Yp(1,1:n);
    // cc=(cvar(2:end)+cvar(1:end-1))/2;
    // cfx=S*sum(fx,1)./dy./cc;
    // cfy=S*sum(fy,1)./dy./cc;
    // cfz=S*sum(fz,1)./dy./cc;
    // Cl=cfz*cos(alpha)-cfx*sin(alpha);
    // Cd=cfx*cos(alpha)+cfz*sin(alpha);