#include "wClosures.h"
#include <cmath>
#include <iostream>

using namespace vii;

// This is an empty constructor
Closures::Closures() {}

Closures::~Closures(){}

/**
 * @brief Laminar closure relations
 */
std::vector<double> Closures::LaminarClosures(double theta, double H, double Ret, double Me, std::string name) const
{

    H = std::max(H, 1.00005);

    /* Kinematic shape factor H* */

    double Hk = (H - 0.29 * Me * Me) / (1 + 0.113 * Me * Me);
    if (name == "wake")
    {
        Hk = std::max(Hk, 1.00005);
    }
    else
    {
        Hk = std::max(Hk, 1.05000);
    }

    /* Density shape parameter */
    
    double Hstar2 = (0.064/(Hk-0.8)+0.251)*Me*Me;

    /* Boundary layer thickness */

    double delta = std::min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta);

    double Hstar = 0.;
    double Hks = Hk - 4.35;
    if (Hk <= 4.35)
    {
        Hstar = 0.0111*Hks*Hks/(Hk+1) - 0.0278*Hks*Hks*Hks/(Hk+1) + 1.528 -0.0002*(Hks*Hk)*(Hks*Hk);
    }
    else
    {
        Hstar = 1.528 + 0.015*Hks*Hks/Hk;
    }
    /* Whitfield's minor additional compressibility correction */
    Hstar = (Hstar + 0.028*Me*Me)/(1+0.014*Me*Me);

    /* Friction coefficient */

    double tmp = 0.;
    double Cf = 0.;
    if (Hk < 5.5)
    {
        tmp = (5.5-Hk)*(5.5-Hk)*(5.5-Hk) / (Hk+1.0);
        Cf = (-0.07 + 0.0727*tmp)/Ret;
    }
    else if (Hk >= 5.5)
    {
        tmp = 1.0 - 1.0/(Hk-4.5);
        Cf = (-0.07 + 0.015*tmp*tmp) /Ret;
    }

    /* Dissipation coefficient */

    double Cd = 0.;
    if (Hk < 4)
    {
        Cd = ( 0.00205*std::pow(4.0-Hk, 5.5) + 0.207 ) * (Hstar/(2*Ret));
    }
    else
    {
        double HkCd = (Hk-4)*(Hk-4);
        double denCd = 1+0.02*HkCd;
        Cd = ( -0.0016*HkCd/denCd + 0.207) * (Hstar/(2*Ret));
    }

    /* Wake relations */
    
    if (name == "wake")
    {
        Cd = 2*(1.10*(1.0-1.0/Hk)*(1.0-1.0/Hk)/Hk)* (Hstar/(2*Ret));
        Cf = 0.;
    }

    double Us = 0.;
    double Cteq = 0.;

    std::vector<double> ClosureVars(8, 0.);

    ClosureVars[0] = Hstar;
    ClosureVars[1] = Hstar2;
    ClosureVars[2] = Hk;
    ClosureVars[3] = Cf;
    ClosureVars[4] = Cd;
    ClosureVars[5] = delta;
    ClosureVars[6] = Cteq;
    ClosureVars[7] = Us;

    return ClosureVars;
}

/**
 * @brief Turbulent closure relations
 */
std::vector<double> Closures::TurbulentClosures(double theta, double H, double Ct, double Ret, double Me, std::string name) const
{
    /*if (std::isnan(Ret))
    {
        std::cout << "Ret is nan. H:" << H << std::endl;
    }*/
    H = std::max(H, 1.00005);
    Ct = std::min(Ct, 0.3);
    //Ct = std::max(std::min(Ct, 0.30), 0.0000001);

    double Hk = (H - 0.29*Me*Me)/(1+0.113*Me*Me);
    if (name == "wake")
    {
        Hk = std::max(Hk, 1.00005);
    }
    else
    {
        Hk = std::max(Hk, 1.05000);
    }

    double Hstar2 = ((0.064/(Hk-0.8))+0.251)*Me*Me;
    double gamma = 1.4 - 1.;
    double Fc = std::sqrt(1+0.5*gamma*Me*Me); /* .2 can be used as well replacing .5*gamma */

    double H0 = 0.;
    if (Ret <= 400)
    {
        H0 = 4.0;
    }
    else
    {
        H0 = 3 + (400/Ret);
    }
    if (Ret <= 200)
    {
        Ret = 200;
    }

    double Hstar = 0.;
    if (Hk <= H0)
    {
        Hstar = ((0.5-4/Ret)*((H0-Hk)/(H0-1))*((H0-Hk)/(H0-1)))*(1.5/(Hk+0.5))+1.5+4/Ret;
    }
    else
    {
        Hstar = (Hk-H0)*(Hk-H0)*(0.007*std::log(Ret)/((Hk-H0+4/std::log(Ret))*(Hk-H0+4/std::log(Ret))) + 0.015/Hk) + 1.5 + 4/Ret;
    }
    /* Whitfield's minor additional compressibility correction */ 
    Hstar = (Hstar + 0.028*Me*Me)/(1+0.014*Me*Me);


    double logRt = std::log(Ret/Fc);
    logRt = std::max(logRt, 3.0);
    double arg = std::max(-1.33*Hk, -20.0);

    /* Equivalent normalized wall slip velocity */
    double Us = (Hstar/2)*(1-4*(Hk-1)/(3*H));

    /* Boundary layer thickness */

    double delta = std::min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta);

    double Ctcon = 0.5/(6.7*6.7*0.75);
    double Hkc = 0.;
    double Cdw = 0.;
    double Cdd = 0.;
    double Cdl = 0.;

    double Hmin = 0.;
    double Fl = 0.;
    double Dfac = 0.;

    double Cf = 0.;
    double Cd = 0.;

    double n = 1.0;

    double Cteq = 0.;
    
    if (name == "wake")
    {
        if (Us > 0.99995)
        {
            Us = 0.99995;
        }
        n = 2.0; // two wake halves
        Cf = 0.0; // no friction inside the wake
        Hkc = Hk - 1;
        Cdw = 0.0;  // wall contribution
        Cdd = (0.995-Us)*Ct*Ct; // turbulent outer layer contribution
        Cdl = 0.15*(0.995-Us)*(0.995-Us)/Ret; // laminar stress contribution to outer layer
        Cteq = std::sqrt(4*Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H)); // Here it is Cteq^0.5
    }
    else
    {
        if (Us > 0.95)
        {
            Us = 0.98;
        }
        n = 1.0;
        Cf = (1/Fc)*(0.3*std::exp(arg)* std::pow(logRt/2.3026, (-1.74-0.31*Hk)) + 0.00011*(std::tanh(4-(Hk/0.875))-1));
        Hkc = std::max(Hk-1-18/Ret, 0.01);
        /* dissipation coefficient */
        Hmin = 1 + 2.1/std::log(Ret);
        Fl = (Hk-1)/(Hmin-1);
        Dfac = 0.5+0.5*std::tanh(Fl);
        Cdw = 0.5*(Cf*Us) * Dfac;
        Cdd = (0.995-Us)*Ct*Ct;
        Cdl = 0.15*(0.995-Us)*(0.995-Us)/Ret;
        Cteq = std::sqrt(Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H)); // Here it is Cteq^0.5
    }
    if (n*Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H)<0)
    {
        std::cout << "Negative sqrt encountered " << std::endl;
        //throw;
    }
    //double Cteq = std::sqrt(n*Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H));
    Cd = n*(Cdw+ Cdd + Cdl);
    // Ueq = 1.25/Hk*(Cf/2-((Hk-1)/(6.432*Hk))**2)

    std::vector<double> ClosureVars(8, 0.);

    ClosureVars[0] = Hstar;
    ClosureVars[1] = Hstar2;
    ClosureVars[2] = Hk;
    ClosureVars[3] = Cf;
    ClosureVars[4] = Cd;
    ClosureVars[5] = delta;
    ClosureVars[6] = Cteq;
    ClosureVars[7] = Us;

    return ClosureVars;
}